#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <limits.h>
#include <ctype.h>

#ifndef ARG_MAX
#define ARG_MAX 10000
#endif

int args;
char *cmd[ARG_MAX + 1];

char *I_flag;
char t_flag;
char r_flag;
char nl_flag;
int n_flag;
int s_flag;

enum {
	NORMAL,
	CEOF,
	ERROR,
	I_FLAG
};

enum {
	NONE,
	QUOTE
};

void clear_cmd(void) {
	for (int i = 0; i < args; i++)
		if (cmd[i] != NULL)
			free(cmd[i]);

	args = 0;
}

int add_arg(const char *str, int chars, int flag) {
	if (args >= ARG_MAX)
		return ERROR;

	else if (n_flag > 0 && args > n_flag && I_flag == NULL)
		return ERROR;

	else if (s_flag > 0 && chars > s_flag)
		return ERROR;

	if (!flag && I_flag) {
		for (int i = 0; i < args; i++)
			if (strstr(cmd[i], I_flag))
				cmd[i] = strdup(str);

		return I_FLAG;
	}

	cmd[args] = strdup(str);
	args++;

	return NORMAL;
}

int is_correct(char c) {
	if (nl_flag)
		return c == '\0';

	else
		return isspace(c);
}

int xargs(void) {
	size_t arg_size = 0;
	char *arg = malloc(1);
	if (arg == NULL) {
		fprintf(stderr, "xargs: malloc failed\n");
		return ERROR;
	}

	size_t index = 0;
	int flag = NONE;
	int ret = NORMAL;
	int chars = 0;
	int args_passed = 0;

	while (1) {
		int c = getchar();
		if (c == EOF) {
			if (flag == QUOTE) {
				fprintf(stderr, "xargs: unterminated quote\n");
				ret = ERROR;
			}

			if (index != arg_size)
				goto PUSH;

			ret = CEOF;
			break;
		}

		if (flag == NONE && is_correct(c) && strlen(arg) > 0) {
			int r = 0;
PUSH:

			r = add_arg(arg, chars, 0);
			if (r == ERROR)
				break;

			else if (r == I_FLAG) {
				free(arg);
				return NORMAL;
			}

			index = 0;
			arg_size = 0;
			args_passed++;

			free(arg);
			arg = malloc(1);
			if (arg == NULL) {
				fprintf(stderr, "xargs: malloc failed\n");
				return ERROR;
			}
		}

		else {
			if (c == '"' || c == '\'') {
				if (flag == QUOTE)
					flag = NONE;

				else
					flag = QUOTE;

				continue;
			}

			arg = realloc(arg, (arg_size++) + 1);
			if (arg == NULL) {
				fprintf(stderr, "xargs: realloc failed\n");
				return ERROR;
			}

			arg[index] = c;

			index++;
			arg[index] = '\0';

			chars++;
		}
	}

	if (args_passed == 0)
		ret = CEOF;

	free(arg);
	return ret;
}

int spawn(void) {
	if (t_flag) {
		for (int i = 0; i < args; i++)
			fprintf(stderr, "%s ", cmd[i]);

		fputc('\n', stderr);
	}

	pid_t pid;
	if ((pid = fork()) == 0) {
		execvp(cmd[0], cmd);
		fprintf(stderr, "xargs: %s\n", strerror(errno));
		exit(1);
	}

	int status = 0;
	waitpid(pid, &status, 0);

	if (status == 255)
		status = 124;

	else if (status >= 0x180)
		status = 125;

	return status;
}

void check(int val) {
	if (val <= 0) {
		fprintf(stderr, "xargs: invalid number: %s\n", optarg);
		exit(1);
	}
}

int main(int argc, char **argv) {
	int opt;
	while ((opt = getopt(argc, argv, "tn:s:rP:0I:")) != -1) {
		switch (opt) {
			case 't':
				t_flag = 1;
				break;

			case 'n':
				n_flag = atoi(optarg);
				check(n_flag);
				break;

			case 's':
				s_flag = atoi(optarg);
				check(s_flag);
				break;

			case 'r':
				r_flag = 1;
				break;

			case '0':
				nl_flag = 1;
				break;

			case 'I':
				I_flag = optarg;
				break;

			default:
				printf("xargs [tnsrP0I] [cmd [arg1] [arg2...]\n\t-t Print the command before start\n\t-n Pass no more than N args\n\t-r Don't run command if input is empty\n\t-s Pass command line of no more than N bytes\n\t-0 NUL terminated input\n\t-I STR Replace STR within PROG ARGS with input line\n");
				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	int ret = 0;
	while (1) {

		/* Arg */
		if (argc) {
			for (int i = 0; i < argc; i++)
				if (add_arg(argv[i], 0, 1))
					break;
		}

		else
			add_arg("echo", 0, 1);

		int stdin_stat = xargs();

		/* Check NULL */
		for (int i = 0; i < args; i++) {
			if (cmd[i] == NULL) {
				fprintf(stderr, "xargs: strdup failed\n");
				clear_cmd();
				return 1;
			}
		}

		/* Run */
		if (stdin_stat == CEOF && I_flag == NULL) {
			if (!r_flag)
				ret = spawn();
		}

		else if (stdin_stat == NORMAL)
			ret = spawn();

		clear_cmd();
		if (stdin_stat)
			break;
	}

	return ret;
}
